//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop


#include <math.h>
#include "CDFT_HIBERT.h"


//---------------------------------------------------------------------------

#pragma package(smart_init)
//---------------------------------------------------
TCDFT_HIBERT::TCDFT_HIBERT()
{
        m_used_flag=false;       
        m_N=0;
       f_used_flag=false;
}
//--------------------------------------------------
TCDFT_HIBERT::~TCDFT_HIBERT()
{
 if( m_used_flag)
 {
    free( m_real_out);
    free( m_real_in);
    free( m_comp_out);
    free( m_comp_in);
    free( h_coeff);
    free( h_freq);
 }

 // old fft subroutine
 f_free();

}
//---------------------------------------------------
bool TCDFT_HIBERT::used_flag()
{

   return m_used_flag;
}
//---------------------------------------------------
bool    TCDFT_HIBERT::get_work_area(int N_span)
{

  m_N= N_span;


  if( m_used_flag)
  {
    free( m_real_out);
    free( m_real_in);
    free( m_comp_out);
    free( m_comp_in);
    free( h_coeff);
    free( h_freq);
  }


  if( (m_real_out=(double *)malloc( (m_N +2 )   * sizeof(double)))==NULL)
  {
    return false;
  }

  if( (m_real_in=(double *)malloc( (m_N +2)   * sizeof(double)))==NULL)
  {
    return false;
  }


  if( (m_comp_out=(double *)malloc( (m_N +2) * 2  * sizeof(double)))==NULL)
  {
    return false;
  }

  if( (m_comp_in=(double *)malloc( (m_N +2) * 2  * sizeof(double)))==NULL)
  {
    return false;
  }

  if( (h_coeff=(double *)malloc( (m_N +2)   * sizeof(double)))==NULL)
  {
    return false;
  }


  if( (h_freq=(double *)malloc( (m_N +2)   * sizeof(double)))==NULL)
  {
    return false;
  }

  //
  m_used_flag=true;
 


  return m_used_flag;
}
//---------------------------------------------------
bool    TCDFT_HIBERT::dft(int iflag, int real_in_flag)
{
     int i,j;
     double y0,y1,x0,xx;

     if( m_used_flag == false ) return false;


     xx= 2.0 * M_PI / (double)m_N;


     for(i=0;i<m_N;++i)     // k
     {
       y0=y1=0.0;
       for(j=0;j<m_N;++j)    // n
       {
          x0= xx * (double)i * (double) j;


         if( real_in_flag == USE_REAL_IN) // use m_real_in as input
         {
            if( iflag == USE_INVERSE)     // inverse
            {
              y0 += cos(x0) * *(m_real_in+j);
              y1 += sin(x0) * *(m_real_in+j);
            }
            else
            {
              y0 += cos(x0) * *(m_real_in+j);
              y1 -= sin(x0) * *(m_real_in+j);
            }
         }
         else  // use m_comp_in as input
         {
            if( iflag == USE_INVERSE)       // inverse
            {



              y0 += cos(x0) * *(m_comp_in + 2 * j) - sin(x0) * *(m_comp_in + 2 * j + 1);
              y1 += sin(x0) * *(m_comp_in + 2 * j) + cos(x0) * *(m_comp_in + 2 * j + 1);
            }
            else
            {
              y0 += cos(x0) * *(m_comp_in + 2 * j) + sin(x0) * *(m_comp_in + 2 * j + 1);
              y1 -= sin(x0) * *(m_comp_in + 2 * j) - cos(x0) * *(m_comp_in + 2 * j + 1);
            }
         }

       } // end of j


       if( iflag == 1) // invrse
       {
         *(m_comp_out + 2 * i) = y0 / (double)m_N;    // real part
         *(m_comp_out + 2 * i + 1 ) =y1 / (double)m_N; // imaginary part

       }
       else
       {
        *(m_comp_out + 2 * i) = y0;    // real part
        *(m_comp_out + 2 * i + 1 ) =y1; // imaginary part
       }

       // just only real part out
       *(m_real_out + i) =  *(m_comp_out + 2 * i);

     } // end of i


     return true;
}
//-------------------------------------
bool    TCDFT_HIBERT::abs_ln(int real_in_flag)
{
     int i;
     double x0;


     if( m_used_flag == false ) return false;

     for(i=0;i< m_N;++i)
     {

        if( real_in_flag == USE_REAL_IN) //  use m_real_in as input
        {
          x0 = m_real_in[i] *  m_real_in[i];
        }
        else        // use m_comp_in as input
        {
          x0 =  m_comp_in[2 * i] *  m_comp_in[2 * i];
          x0 += m_comp_in[2 * i + 1] *  m_comp_in[2 * i + 1];
        }

        if( x0 <= 0.0)
        {
          // over-write m_real_in
          m_real_in[i] = -1.0E-200;     // set dummy small data

          // return false;
        }else
        {
        // over-write m_real_in
          m_real_in[i] = log(sqrt(x0));    // log is ln
          m_comp_in[2*i]= m_real_in[i];
          m_comp_in[2*i+1]=0.0;
        }
     }


     return true;
}
//-------------------------------------
bool    TCDFT_HIBERT::mk_hibert_filter()
{
     int i;
     double x0;


     if( m_used_flag == false ) return false;


     // set all zero
     for(i=0;i< m_N;++i)
     {
       h_coeff[i]=0;
     }

     // 1st is always 1
     h_coeff[0]=1.0;

     //
     if( (m_N % 2) == 0)
     {
         h_coeff[ m_N / 2 ] = 1.0;
         for(i=1;i< (m_N/2);++i)  h_coeff[i] = 2.0;
     }
     else
     {
          for(i=1;i< ((m_N+1)/2);++i)  h_coeff[i] = 2.0;
     }



     return true;
}
//-------------------------------------
bool    TCDFT_HIBERT::cal_freq_resp(int nfir0, double *fir_coeff)
{
      int i,j;
      double x0,y0,xw,dw0;

      dw0=  2.0 * M_PI / (double)m_N;

      for(i=0;i<m_N;++i)
      {
         xw=  dw0 * (double)i;
         x0=y0=0.0;

         for(j=0;j<nfir0;++j)
         {
            x0 += fir_coeff[j] * cos( xw * (double)j);
            y0 -= fir_coeff[j] * sin( xw * (double)j);
         }

         m_comp_in[2*i] = x0;  // real part
         m_comp_in[2*i + 1]= y0; // imag part

         h_freq[i]= sqrt(x0 * x0 + y0 * y0);      // abs freqeucny response

      }



      return true;
}
//-------------------------------------
bool    TCDFT_HIBERT::do_hibert(int fs, int nfir0, double *fir_coeff)
//    DFT vZɎԂ
//
// suppose that  input fir coeff
//               calculate      frequency response length is as same as fs
//               output impluse response length is as same as fs.
//
{
      int i;
      double x0;


      get_work_area(fs);
      cal_freq_resp(nfir0, fir_coeff);
      mk_hibert_filter();       // prepare hilbert filter coefficeint

      abs_ln( USE_COMP_IN);     // ln(abs(input frequecny response)



      dft( USE_INVERSE, USE_REAL_IN);

      for(i=0;i< m_N;++i)
      {
        m_comp_in[2 * i]= m_comp_out[2 * i] * h_coeff[i];
        m_comp_in[2 * i + 1]= m_comp_out[2 * i + 1] * h_coeff[i];
      }


      dft( USE_NON_INVERSE, USE_COMP_IN);      //   out imaginary part  is approximate minimum phase (only phase information)

      for(i=0;i< m_N;++i)
      {
        x0=  m_comp_out[2 * i+1];  // imagnary part
        m_comp_in[2 * i]= h_freq[i] * cos( x0);
        m_comp_in[2 * i + 1]=  h_freq[i] * sin(x0);
      }

      dft( USE_INVERSE, USE_COMP_IN);     // out is complex impluse response


      // vZꂽCpX̌ʂ́Am_real_out[] Ɋi[ĂB



     return true;
}
//-------------------------------------
//
//  Below is old fft subroutine
//
//----------------------------------------------
bool TCDFT_HIBERT::f_set( int fs)
{
//    fs ɋ߂@fs菬@2ׂ̂Ԃ
//
     nzyou0 = (int) (log((double)fs)/log(2.0));

     len0=f_zyou(nzyou0);
     f_iv_mkdat( nzyou0);
     f_mkdat( nzyou0 );
     f_bitrev( nzyou0);


     f_used_flag=true;

     return true;
}
//---------------------------------------------
bool TCDFT_HIBERT::f_free()
{

    if( f_used_flag)
    {
      free(mk);
      free(ivmk);
      free(table);
    }
     return true;
}
//----------------------------------------------
int TCDFT_HIBERT::f_zyou( int zyou)
{
   if( zyou < 0)
   {
     return(1);
   }
   else return( 1 << zyou);
}
//----------------------------------------------------------------------------
int TCDFT_HIBERT::f_mkdat( int zyou )
{
   int i,z;
   double x;
   double *p,*p0;

   z= f_zyou(zyou);
   if( (p = (double *)malloc( sizeof(double) * z *2 )  ) == NULL)
   {
       return(-1);
   }
    p0=p;
   for(i=0;i<z;++i)
   {
       x= M_PI * 2.0 *(double)i / (double)z;
       *p= cos(x);
       ++p;
       *p= -sin(x);
       ++p;
   }

   //
   mk=p0;
   return(0);
}
//-----------------------------------------------------------------
int TCDFT_HIBERT::f_iv_mkdat( int zyou )
/*  inverted data  */
/*  for i-fft             */
{
   int i,z;
   double x;
   double *p,*p0;

   z= f_zyou(zyou);
   if( (p = (double *)malloc( sizeof(double) * z *2 )  ) == NULL)
   {
      return(-1);
  }
    p0=p;
   for(i=0;i<z;++i)
   {
       x= M_PI * 2.0 *(double)i / (double)z;
       *p= cos(x);
       ++p;
       *p= sin(x);
       ++p;
   }
   //
   ivmk=p0;
   return(0);
}
//----------------------------------------------------------------
int TCDFT_HIBERT::f_bitrev(int nzyou)
/* data[0]- data[ 2^^nzyou-1]= 2^^nzyou ko   */
{
   int i,j,x1,x;
   int out;
   int kosuu;
   int *p,*p0;

   kosuu=f_zyou(nzyou);

  if( (p=(int *)malloc( sizeof(int) * kosuu)) == NULL)
   {
      return(-1);
  }
   p0=p;
   for(i=0;i<kosuu;++i)
   {
     out=0;
     x1=i;
     for(j=nzyou;j>0;--j)
     {
        x = x1 & 0x0001;
        out = out + x*f_zyou(j-1);
        x1>>=1;
     }
     *p=out;
     ++p;
   }
   //
   table=p0;
   return(0);
}


//-----------------------------------------------
int TCDFT_HIBERT::f_fft(double *datas, int mode)
{                                                    
/*  data[2*i]= real-part; data[1+2*i]= imaginaly-part                 */
/* kono subroutone wo yobumaeni ws to table wo set suru koto          */
/*  *******************************************************************/
/*  inverse-fft mode==3, else mode==0           */

   int n,nh,i,i1,i2,x1,x2,offset;
   double *ws,a[2],bn1;


   /* start fft */
   nh=len0;
   bn1= 1.0/(double)nh;

   if( mode == 3)
   {
     ws= ivmk;   //
   }
   else ws=mk;

   for(i1=0;i1<nzyou0;++i1)
   {
         nh= nh/2;
         n= f_zyou(i1);

         for(i2=0;i2<n;++i2)
         {
            offset=2*nh*i2;
            for(i=0;i<nh;++i)
            {
               x1=2*(i+offset);
               x2=2*(i+nh+offset);
               a[0]= datas[x1] - datas[x2];
               a[1]= datas[x1+1] - datas[x2+1];
               datas[x1]   = datas[x1] + datas[x2];
               datas[x1+1] = datas[x1+1] + datas[x2+1];
               datas[x2]   = a[0] * ws[2*i*n]   - a[1] * ws[2*i*n+1];
               datas[x2+1] = a[0] * ws[2*i*n+1] + a[1] * ws[2*i*n];
             }
         }
   }

   /* iff 1/N wo kakeru */
   if( mode == 3)
   {
     for(i=0;i<len0;++i)
     {
       datas[2*i]   = datas[2*i] * bn1;
       datas[2*i+1] = datas[2*i+1] * bn1;
     }
   }



   /* irekae */

   for(i=0;i<len0;++i)
   {
     i1= table[i];

     if( i < i1)
     {
      a[0] = datas[2*i1];
      a[1] = datas[2*i1+1];

      datas[2*i1]=datas[2*i];
      datas[2*i1+1]=datas[2*i+1];

      datas[2*i]   = a[0];
      datas[2*i+1] = a[1];
     }

    }


   return(0);
}
//------------------------------------------------------------------
bool    TCDFT_HIBERT::do_hibert2(int fs, int nfir0, double *fir_coeff)
//        FFT Qׂ̂ߎvZ
//
// suppose that  input fir coeff
//               calculate      frequency response length is as same as fs
//               output impluse response length is as same as fs.
//
{
      int i;
      double x0,y0;


      f_set(fs);

      get_work_area(len0);

      for(i=0;i< m_N;++i)
      {
        if( i< nfir0)
        {
          m_comp_in[2 * i]= fir_coeff[i];
        }
        else  { m_comp_in[2 * i]=0.0;}
        m_comp_in[2 * i + 1]= 0.0;
      }
      f_fft(m_comp_in, 0);


      for(i=0;i< m_N;++i)
      {
         x0=  m_comp_in[2*i];  // real part
         y0=  m_comp_in[2*i + 1];; // imag part
         h_freq[i]= sqrt(x0 * x0 + y0 * y0);      // abs freqeucny response
      }
      

      mk_hibert_filter();       // prepare hilbert filter coefficeint

      abs_ln( USE_COMP_IN);     // ln(abs(input frequecny response)

      f_fft(m_comp_in, 3);

      for(i=0;i< m_N;++i)
      {

        m_comp_in[2 * i]= m_comp_in[2 * i] * h_coeff[i];
        m_comp_in[2 * i + 1]= m_comp_in[2 * i + 1] * h_coeff[i];
      }


       f_fft(m_comp_in, 0);

      for(i=0;i< m_N;++i)
      {
        x0=  m_comp_in[2 * i+1];  // imagnary part
        m_comp_in[2 * i]= h_freq[i] * cos( x0);
        m_comp_in[2 * i + 1]=  h_freq[i] * sin(x0);
      }

      f_fft(m_comp_in, 3);


 // vZꂽCpX̌ʂ́Am_comp_out[2*i] i=0,1,2,....Ɋi[ĂB


     return true;
}
//------------------------------------------------------------
